/*
 * bf_brute3
 * by Javantea aka Joel R. Voss
 * Sept 23, 2006
 * Attack weak keys on symmetric ciphers.
 */

#include <string.h>
#include "check_tar.h"
#include "increment.h"
#include <ssl_bf2a.h>

// One global variable.
int verbose = 0;

int decrypt_bfk(const BYTE *enc_data, int enc_size, BYTE *key, BYTE *iv, BYTE *out, int *data_size)
{
	EVP_CIPHER_CTX ctx;
	BYTE           salt[PKCS5_SALT_LEN];

	const char *ptr_enc;
	char  final[EVP_MAX_BLOCK_LENGTH];
	int   ret = 0;

    // Take the encrypted data from the stdin.
	if(0 == strncmp(enc_data, BF_HEADER, 8))
	{
		memcpy(salt, enc_data + 8, PKCS5_SALT_LEN);
    
	//generate_bin_key(key, password, pass_len, salt, iv, KEY_LENGTH);
    //print_hex(key, KEY_LENGTH);
		ptr_enc = enc_data + 8 + PKCS5_SALT_LEN;
    // Subtract the header from the encrypted size.
		enc_size -= (8 + PKCS5_SALT_LEN);
	}
	else
	{
		if(verbose) fprintf(stderr, "Warning: BF Decrypt Header not found: %08x.\n", *(unsigned int *)enc_data);
		//return 1;
		ptr_enc = enc_data;
		
	}
	EVP_CIPHER_CTX_init(&ctx);
	EVP_DecryptInit(&ctx, EVP_bf_cbc(), key, iv);
	char dec_data[8+EVP_CIPHER_CTX_block_size(&ctx)];
	int j, k = 0;
	for(j=0; j<enc_size; j+=8)
	{
		int outl = decrypt_data(&ctx, dec_data, ptr_enc + j, 8);
        // The first one is always null, ignore it or fix it.
        // TODO: You should probably try to understand it as well.
		if(outl != 0)
		{
			memcpy(out + k, dec_data, 8);
			//free(dec_data);
			// Counter for output.
			k += 8;
		}
		//else
		//	fprintf(stderr, "Error: Decrypt failed, null data.\n");
	}
	if (!EVP_DecryptFinal(&ctx, final, data_size))
	{
		if(verbose) fprintf(stderr, "Error: Decrypt Final, padding incorrect.\n");
		ret = 1;
        //abort();
	}
	EVP_CIPHER_CTX_cleanup(&ctx);
	memcpy(out + k, final, *data_size);
    // Data size is set in DecryptFinal to the last size (0-7).
    // Add k (output counter) to get the correct total size.
	*data_size += k;

	return ret;
}

int bf_brute_key(const char *file_data, int size, BYTE *key, BYTE *iv)
{
	// 1) Decrypt file_data with password.
	char out_data[size];
	int out_size = -1;
	int ret = decrypt_bfk(file_data, size, key, iv, out_data, &out_size);
	// 2) check_tar(out_data, out_size)
	return check_tar(out_data, out_size);
}

#ifdef __BF_BRUTE3_SA__

int hex_input(BYTE *output, const char *input, int len)
{
	int size = strlen(input);
	int i, j = 0, hex;
	// Don't allow an attempted overflow.
	if(size > 2 * len)
	{
		size = 2 * len;
	}
	for(i = 0; i < size; i++)
	{
		if(input[i] >= '0' && input[i] <= '9')
		{
			hex = input[i] - '0';
		}
		else if(input[i] >= 'a' && input[i] <= 'f')
		{
			hex = 10 + input[i] - 'a';
		}
		else if(input[i] >= 'A' && input[i] <= 'F')
		{
			hex = 10 + input[i] - 'A';
		}
		else
		{
			// End of matches.
			return i;
		}
		output[j] = hex << 4;
		i++;
		if(input[i] >= '0' && input[i] <= '9')
		{
			hex = input[i] - '0';
		}
		else if(input[i] >= 'a' && input[i] <= 'f')
		{
			hex = 10 + input[i] - 'a';
		}
		else if(input[i] >= 'A' && input[i] <= 'F')
		{
			hex = 10 + input[i] - 'A';
		}
		else
		{
			// End of matches.
			return i;
		}
		output[j] |= hex;
		j++;
	}
	return i;
}

void usage()
{
	printf("Usage: bf_brute3 [-v] file.tar.bf 00000000000000000000000000000000\n");
}

int main(int argc, char **argv)
{
	if(argc < 2 || (strncmp(argv[1], "-h", 2) == 0))
	{
		usage();
		if(argc >= 3)
		{
			BYTE output[16] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
			int ret = hex_input(output, argv[2], 16);
			fprintf(stderr, "%i = hex_input(\"%s\", ", ret, argv[2]);
			print_hex(output, 16);
			fprintf(stderr, ", 16);\n");
		}
		return 1;
	}
	int current_arg = 1;
	if((strncmp(argv[current_arg], "-v", 2) == 0))
	{
		verbose = strlen(argv[current_arg]) - 1;
		current_arg++;
	}
	// Attack a file to find the password.
	char indata[256];
	FILE *infile = fopen(argv[current_arg], "rb");
	if(infile == 0)
	{
		printf("File is invalid: %s\n", argv[1]);
		usage();
		return 1;
	}
	current_arg++;
	int size = fread(indata, 1, 256, infile);
	fclose(infile);
	BYTE key[KEY_LENGTH] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
	BYTE iv[EVP_MAX_IV_LENGTH] = "\0\0\0\0";
	if(argc >= 3)
	{
		hex_input(key, argv[current_arg], KEY_LENGTH);
	}
	fprintf(stderr, "Starting with key: ");
	print_hex(key, 16);
	int i = 0;
	while(1)
	{
		if(bf_brute_key(indata, size, key, iv) == 0)
		{
			fprintf(stderr, "Possible key found: ");
			print_hex(key, KEY_LENGTH);
			fprintf(stderr, "\n");
			return 0;
		}
		increment(key, KEY_LENGTH, key, 0, 0xff, 0);
		if(i++ >= 30000)
		{
			i = 0;
			print_hex(key, 16);
			//sleep(1);
		}
	}
	return 1;
}
#endif // __BF_BRUTE3_SA__
